// Nodelist support
// by Przemyslaw Kwiatkowski, Warsaw, Poland, 2:480/127, micha@fidonet.org.pl

/*
 * To do: (any volunteers?)
 *
 * - nodelist indexing (maybe V7+ support?)
 * - domain support
 * - improving speed
 * - debugging :-)
 * - ...any ideas?
 */

class HubInfo {
    public:
        word  zone, net, node;
        word *downlinks;
        ULONG linksTotal;

        int addLink(word);
        HubInfo(word, word, word);
        ~HubInfo();
};

HubInfo::HubInfo(word x, word y, word z) {
        zone=x;
        net=y;
        node=z;
        linksTotal=0;
        downlinks=NULL;
}

HubInfo::~HubInfo() {
    if (linksTotal>0) free(downlinks);
}

int HubInfo::addLink(word node) {
        if ((downlinks=(word*)realloc(downlinks, sizeof(word)*++linksTotal))==NULL) return NOMEMORY;
        downlinks[linksTotal-1]=node;
    return SUCCESS;
}



class HubsDatabase {
    private:
        HubInfo **hubs; // not sorted (so slow...)
        ULONG     hubsTotal;
    public:
        HubsDatabase();
        ~HubsDatabase();
        HubInfo *findHub(word, word, word);
        int addHub(word, word, word);  // zone, net, node
        int addDownlink(word);           // node# only, will be placed under the *last* Hub
};

HubsDatabase::HubsDatabase() {
        hubs=NULL;
        hubsTotal=0;
}

HubsDatabase::~HubsDatabase() {
        while(hubsTotal>0) delete(hubs[--hubsTotal]);
}

int HubsDatabase::addHub(word zone, word net, word node) {
        if ((zone==0)||(net==0)) {
                printf(" - Error!\n");
                return FATAL;
        }
        if (findHub(zone, net, node)!=NULL) {
        printf(" - Error: Duplicated Hub!\n");
                return FATAL;
        }
        if ((hubs=(HubInfo**)realloc(hubs, sizeof(HubInfo*)*++hubsTotal))==NULL) return NOMEMORY;
        if ((hubs[hubsTotal-1]=new HubInfo(zone, net, node))==NULL) return NOMEMORY;
    return SUCCESS;
}

int HubsDatabase::addDownlink(word node) {
        if (hubsTotal==0) {
        printf("- Error!\n");
                return FATAL;
        }
    return hubs[hubsTotal-1]->addLink(node);
}

HubInfo *HubsDatabase::findHub(word x, word y, word z) { // slow, but works fine :-)
        for(ULONG i=0; i<hubsTotal; ++i)
                if ((hubs[i]->zone==x)&&(hubs[i]->net==y)&&(hubs[i]->node==z)) return (hubs[i]);
    return NULL;
}


class HubsDatabase Hubs;

int ProcParNodelist (char *Token, struct S_FQAddress) {
    CConfigHandler Nodelist;
    char Line[256];
    word zone, net, node;
    struct ffblk ffblk;
    int done, maxtime, maxdate;
    char name[256], fullname[512], *s, *t;
    unsigned hubscount=0, nodescount=0;

        // looking for the newest nodelist (filename.###)
        // optimalization of these code would be a good idea :-)
        for(s=Token; (t=strchr(s+1, DIRSEPC))!=NULL; s=t);
        strcpy(name, Token);
        strcat(name, ".???");
        if ((done=findfirst(name,&ffblk,0))==-1) {
                printf ("Unable to open nodelist %s\n", Token);
                return ECH_OPENFAILED;
        }
        while (!((isdigit(ffblk.ff_name[strlen(ffblk.ff_name)-1])&&isdigit(ffblk.ff_name[strlen(ffblk.ff_name)-2])&&isdigit(ffblk.ff_name[strlen(ffblk.ff_name)-3]))||done)) {
                done=findnext(&ffblk);
        }
        if (done!=0) {
        printf ("Unable to open nodelist %s\n", Token);
                return ECH_OPENFAILED;
        }

        strcpy(name, ffblk.ff_name);
        maxdate=ffblk.ff_fdate;
        maxtime=ffblk.ff_ftime;
        while (!done) {
                done=findnext(&ffblk);
                if ((ffblk.ff_fdate>maxdate)||((ffblk.ff_fdate==maxdate)&&(ffblk.ff_ftime>maxtime))&&isdigit(ffblk.ff_name[strlen(ffblk.ff_name)-1])&&isdigit(ffblk.ff_name[strlen(ffblk.ff_name)-2])&&isdigit(ffblk.ff_name[strlen(ffblk.ff_name)-3])) {
                        strcpy(name, ffblk.ff_name);
                        maxdate=ffblk.ff_fdate;
                        maxtime=ffblk.ff_ftime;
                }
        }
        findclose(&ffblk);
        
        *s=0;
        strcpy(fullname, Token);
        if (*fullname==0) strcat(fullname, ".");
        strcat(fullname, DIRSEPS);
        strcat(fullname, name);

        
        if (Nodelist.Open(fullname)!=SUCCESS) {
                printf ("Unable to open nodelist %s\n",fullname);
                return ECH_OPENFAILED;
        }
        printf("Reading nodelist: %s ", Nodelist.ConfigFileName);
        
        while (Nodelist.ReadLine(Line)!=ECH_EOF) {
                if (!strnicmp(Line, "Zone,", 5)) {
                        zone=net=atoi(Line+5);
                        if (Hubs.addHub(zone,net,0)!=SUCCESS) return FATAL;
                        ++hubscount;
                } else if (!strnicmp(Line, "Region,", 7)) {
                        net=atoi(Line+7);
                        if (Hubs.addHub(zone,net,0)!=SUCCESS) return FATAL;
                        ++hubscount;
                } else if (!strnicmp(Line, "Host,", 5)) {
                        net=atoi(Line+5);
                        if (Hubs.addHub(zone,net,0)!=SUCCESS) return FATAL;
                        ++hubscount;
                } else if (!strnicmp(Line, "Hub,", 4)) {
                        node=atoi(Line+4);
                        if (Hubs.addHub(zone,net,node)!=SUCCESS) return FATAL;
                        ++hubscount;
                } else if ((!strnicmp(Line, "Hold,", 5)) || (!strnicmp(Line, "Down,", 5))) {
                        if ((node=atoi(Line+5))==0) return FATAL;
                        if (Hubs.addDownlink(node)!=SUCCESS) return FATAL;
                        ++nodescount;
                } else if (!strnicmp(Line, "Pvt,", 4)) {
                        if ((node=atoi(Line+4))==0) return FATAL;
                        if (Hubs.addDownlink(node)!=SUCCESS) return FATAL;
                        ++nodescount;
                } else if (*Line==',') {
                        if ((node=atoi(Line+1))==0) return FATAL;
                        if (Hubs.addDownlink(node)!=SUCCESS) return FATAL;
                        ++nodescount;
                } else if (!strnicmp(Line, "Point,", 6)) {
                        // nothing here :-)
                } else if (*Line==26) {
                        break;
                } else {
                        printf("- Error!\n");
                        return FATAL;
                }
        }

    printf("- Hubs: %u   Nodes: %u\n", hubscount, nodescount);
    return SUCCESS;
}



int ProcessHub (char *Hub, int (*CommandProcessor) (char *,S_FQAddress), struct S_FQAddress Main)
{
    S_FQAddress addr;
    HubInfo *hub;
    char tmp[30];

        if (GetFullQualifiedAddress(Hub, &addr, &Main)!=SUCCESS) {
                printf("Warning: Unable to recognize hub address: %s\n", Hub);
                return WARNING;
        }

        if (addr.Point!=0) {
                printf("Warning: Point cannot be hub!\n");
                sprintf(tmp, "%d:%d/%d.%d", addr.Zone, addr.Net, addr.Node, addr.Point);
                if (CommandProcessor(tmp, Main)!=SUCCESS) return FATAL;
                return WARNING;
        }

        sprintf(tmp, "%d:%d/%d.*", addr.Zone, addr.Net, addr.Node);
        if (CommandProcessor(tmp, Main)!=SUCCESS) {
                printf ("Fatal error processing hub %s.\n", Hub);
                return FATAL;
        }

        if ((hub=Hubs.findHub(addr.Zone, addr.Net, addr.Node))==NULL) {
                printf("Warning: No such hub: %d:%d/%d\n", addr.Zone, addr.Net, addr.Node);
                return WARNING;
        }

        for(unsigned i=0; i<(hub->linksTotal); ++i) {
                sprintf(tmp, "%d:%d/%d.*", hub->zone, hub->net, hub->downlinks[i]);
                if (CommandProcessor(tmp, Main)!=SUCCESS) {
                        printf ("Fatal error processing hub %s.\n", Hub);
                        return FATAL;
                }
        }
    return SUCCESS;
}
